#!/usr/bin/env python
# License: GPLv3 Copyright: 2013, Kovid Goyal <kovid at kovidgoyal.net>

import os
from functools import partial
from itertools import product

from qt.core import (
    QAction,
    QDockWidget,
    QEvent,
    QHBoxLayout,
    QIcon,
    QLabel,
    QMenu,
    QMenuBar,
    QSize,
    QStackedWidget,
    Qt,
    QTabWidget,
    QTimer,
    QUrl,
    QVBoxLayout,
    QWidget,
    pyqtSignal,
)

from calibre import prints
from calibre.constants import DEBUG, __appname__, get_version, ismacos
from calibre.customize.ui import find_plugin
from calibre.gui2 import elided_text, open_url, qapplication_or_fail
from calibre.gui2.keyboard import Manager as KeyboardManager
from calibre.gui2.main_window import MainWindow
from calibre.gui2.throbber import ThrobbingButton
from calibre.gui2.tweak_book import actions, capitalize, current_container, editors, toolbar_actions, tprefs, update_mark_text_action
from calibre.gui2.tweak_book.boss import Boss
from calibre.gui2.tweak_book.char_select import CharSelect
from calibre.gui2.tweak_book.check import Check
from calibre.gui2.tweak_book.check_links import CheckExternalLinks
from calibre.gui2.tweak_book.editor.insert_resource import InsertImage
from calibre.gui2.tweak_book.editor.widget import register_text_editor_actions
from calibre.gui2.tweak_book.file_list import FileListWidget
from calibre.gui2.tweak_book.function_replace import DebugOutput
from calibre.gui2.tweak_book.job import BlockingJob
from calibre.gui2.tweak_book.live_css import LiveCSS
from calibre.gui2.tweak_book.manage_fonts import ManageFonts
from calibre.gui2.tweak_book.plugin import create_plugin_actions, install_plugin
from calibre.gui2.tweak_book.preview import Preview
from calibre.gui2.tweak_book.reports import Reports
from calibre.gui2.tweak_book.search import SavedSearches, SearchPanel
from calibre.gui2.tweak_book.spell import SpellCheck
from calibre.gui2.tweak_book.text_search import TextSearch
from calibre.gui2.tweak_book.toc import TOCViewer
from calibre.gui2.tweak_book.undo import CheckpointView
from calibre.gui2.widgets2 import MessagePopup
from calibre.utils.icu import ord_string, sort_key
from calibre.utils.localization import _, localize_user_manual_link, localize_website_link, pgettext
from calibre.utils.unicode_names import character_name_from_code


def open_donate():
    open_url(QUrl(localize_website_link('https://calibre-ebook.com/donate')))


class Central(QStackedWidget):  # {{{
    "The central widget, hosts the editors"

    current_editor_changed = pyqtSignal()
    close_requested = pyqtSignal(object)

    def __init__(self, parent=None):
        QStackedWidget.__init__(self, parent)
        self.welcome = w = QLabel('<p>' + _('Double click a file in the left panel to start editing it.'))
        self.addWidget(w)
        w.setWordWrap(True)
        w.setAlignment(Qt.AlignmentFlag.AlignTop | Qt.AlignmentFlag.AlignHCenter)

        self.container = c = QWidget(self)
        self.addWidget(c)
        l = QVBoxLayout(c)
        c.setLayout(l)
        l.setContentsMargins(0, 0, 0, 0)
        self.editor_tabs = t = QTabWidget(c)
        l.addWidget(t)
        t.setDocumentMode(True)
        t.setTabsClosable(True)
        t.setMovable(True)
        self.modified_icon = QIcon.ic('modified.png')
        self.editor_tabs.currentChanged.connect(self.current_editor_changed)
        self.editor_tabs.tabCloseRequested.connect(self._close_requested)
        self.search_panel = SearchPanel(self)
        l.addWidget(self.search_panel)
        self.restore_state()
        tab_bar = self.editor_tabs.tabBar()
        assert tab_bar is not None
        tab_bar.installEventFilter(self)

    def _close_requested(self, index):
        editor = self.editor_tabs.widget(index)
        self.close_requested.emit(editor)

    def add_editor(self, name, editor):
        fname = name.rpartition('/')[2]
        index = self.editor_tabs.addTab(editor, fname)
        self.editor_tabs.setTabToolTip(index, _('Full path:') + ' ' + name)
        editor.modification_state_changed.connect(self.editor_modified)

    @property
    def tab_order(self):
        ans = []
        rmap = {v: k for k, v in editors.items()}
        for i in range(self.editor_tabs.count()):
            name = rmap.get(self.editor_tabs.widget(i))
            if name is not None:
                ans.append(name)
        return ans

    def rename_editor(self, editor, name):
        for i in range(self.editor_tabs.count()):
            if self.editor_tabs.widget(i) is editor:
                fname = name.rpartition('/')[2]
                self.editor_tabs.setTabText(i, fname)
                self.editor_tabs.setTabToolTip(i, _('Full path:') + ' ' + name)

    def show_editor(self, editor):
        self.setCurrentIndex(1)
        self.editor_tabs.setCurrentWidget(editor)

    def close_editor(self, editor):
        for i in range(self.editor_tabs.count()):
            if self.editor_tabs.widget(i) is editor:
                self.editor_tabs.removeTab(i)
                if self.editor_tabs.count() == 0:
                    self.setCurrentIndex(0)
                return True
        return False

    def editor_modified(self, *args):
        tb = self.editor_tabs.tabBar()
        assert tb is not None
        for i in range(self.editor_tabs.count()):
            editor = self.editor_tabs.widget(i)
            modified = getattr(editor, 'is_modified', False)
            tb.setTabIcon(i, self.modified_icon if modified else QIcon())

    def close_current_editor(self):
        ed = self.current_editor
        if ed is not None:
            self.close_requested.emit(ed)

    def close_all_but_current_editor(self):
        self.close_all_but(self.current_editor)

    def close_to_right_of_current_editor(self):
        self.close_to_right(self.current_editor)

    def close_all_but(self, ed):
        close = []
        if ed is not None:
            for i in range(self.editor_tabs.count()):
                q = self.editor_tabs.widget(i)
                if q is not None and q is not ed:
                    close.append(q)
        for q in close:
            self.close_requested.emit(q)

    def close_to_right(self, ed):
        close = []
        if ed is not None:
            found = False
            for i in range(self.editor_tabs.count()):
                q = self.editor_tabs.widget(i)
                if found:
                    close.append(q)
                elif q is ed:
                    found = True
        for q in close:
            self.close_requested.emit(q)

    @property
    def current_editor(self):
        return self.editor_tabs.currentWidget()

    def save_state(self):
        tprefs.set('search-panel-visible', self.search_panel.isVisible())
        self.search_panel.save_state()
        for ed in editors.values():
            ed.save_state()
        if self.current_editor is not None:
            self.current_editor.save_state()  # Ensure the current editor saves it state last

    def restore_state(self):
        self.search_panel.setVisible(tprefs.get('search-panel-visible', False))
        self.search_panel.restore_state()

    def show_find(self):
        self.search_panel.show_panel()

    def pre_fill_search(self, text):
        self.search_panel.pre_fill(text)

    def eventFilter(self, a0, a1):
        base = super()
        if (
            a0 is not self.editor_tabs.tabBar()
            or a1.type() != QEvent.Type.MouseButtonPress
            or a1.button() not in (Qt.MouseButton.RightButton, Qt.MouseButton.MiddleButton)
        ):
            return base.eventFilter(a0, a1)
        tab_bar = self.editor_tabs.tabBar()
        assert tab_bar is not None
        index = tab_bar.tabAt(a1.pos())
        if index < 0:
            return base.eventFilter(a0, a1)
        if a1.button() == Qt.MouseButton.MiddleButton:
            self._close_requested(index)
        ed = self.editor_tabs.widget(index)
        if ed is not None:
            menu = QMenu(self)
            menu.addAction(actions['close-current-tab'].icon(), _('Close tab'), partial(self.close_requested.emit, ed))
            menu.addSeparator()
            menu.addAction(actions['close-all-but-current-tab'].icon(), _('Close other tabs'), partial(self.close_all_but, ed))
            menu.addAction(
                actions['close-tabs-to-right-of'].icon(),
                _('Close tabs to the right of this tab'),
                partial(self.close_to_right, ed),
            )
            menu.exec(tab_bar.mapToGlobal(a1.pos()))

        return True


# }}}


class CursorPositionWidget(QWidget):  # {{{
    def __init__(self, parent):
        QWidget.__init__(self, parent)
        self.l = QHBoxLayout(self)
        self.setLayout(self.l)
        self.la = QLabel('')
        self.l.addWidget(self.la)
        self.l.setContentsMargins(0, 0, 0, 0)
        f = self.la.font()
        f.setBold(False)
        self.la.setFont(f)

    def update_position(self, line=None, col=None, character=None):
        if line is None:
            self.la.setText('')
        else:
            try:
                name = character_name_from_code(ord_string(character)[0]) if character and tprefs['editor_show_char_under_cursor'] else None
            except Exception:
                name = None
            text = _('Line: {0} : {1}').format(line, col)
            if not name:
                name = {'\t': 'TAB'}.get(character, None)
            if name and tprefs['editor_show_char_under_cursor']:
                text = name + ' : ' + text
            self.la.setText(text)


# }}}


def install_new_plugins():
    from calibre.utils.config import JSONConfig

    prefs = JSONConfig('newly-installed-editor-plugins')
    pl = prefs.get('newly_installed_plugins', ())
    if pl:
        for name in pl:
            plugin = find_plugin(name)
            if plugin is not None:
                install_plugin(plugin)
        prefs['newly_installed_plugins'] = []


class Main(MainWindow):
    APP_NAME = _('Edit book')
    STATE_VERSION = 0
    undo_requested = pyqtSignal(object)

    def __init__(self, opts, notify=None):
        MainWindow.__init__(self, opts, disable_automatic_gc=True)
        self.message_popup = MessagePopup(self)
        self.message_popup.undo_requested.connect(self.undo_requested)
        try:
            install_new_plugins()
        except Exception:
            import traceback

            traceback.print_exc()
        self.setWindowTitle(self.APP_NAME)
        self.boss = Boss(self, notify=notify)
        if not ismacos:
            self.setWindowIcon(qapplication_or_fail().windowIcon())
        self.opts = opts
        self.path_to_ebook = None
        self.container = None
        self.current_metadata = None
        self.blocking_job = BlockingJob(self)
        self.keyboard = KeyboardManager(self, config_name='shortcuts/tweak_book')

        self.central = Central(self)
        self.setCentralWidget(self.central)
        self.check_book = Check(self)
        self.spell_check = SpellCheck(parent=self)
        self.toc_view = TOCViewer(self)
        self.text_search = TextSearch(self)
        self.saved_searches = SavedSearches(self)
        self.image_browser = InsertImage(self, for_browsing=True)
        self.reports = Reports(self)
        self.check_external_links = CheckExternalLinks(self)
        self.insert_char = CharSelect(self)
        self.manage_fonts = ManageFonts(self)
        self.sr_debug_output = DebugOutput(self)

        self.create_actions()
        self.create_toolbars()
        self.create_docks()
        self.create_menubar()

        self.status_bar = self.statusBar()
        assert self.status_bar is not None
        self.status_bar.addPermanentWidget(self.boss.save_manager.status_widget)
        self.cursor_position_widget = CursorPositionWidget(self)
        self.status_bar.addPermanentWidget(self.cursor_position_widget)
        v = get_version()
        self.status_bar_default_msg = la = QLabel(' ' + _('{0} {1} created by {2}').format(__appname__, v, 'Kovid Goyal'))
        setattr(la, 'base_template', str(la.text()))
        setattr(la, 'editing_template', _('{appname} {version} editing: {{path}}').format(appname=__appname__, version=v))
        self.status_bar.addWidget(la)

        self.boss(self)
        screen = self.screen()
        assert screen is not None
        g = screen.availableSize()
        self.resize(g.width() - 50, g.height() - 50)

        self.apply_settings()
        QTimer.singleShot(0, self.restore_state)

    def apply_settings(self):
        self.keyboard.finalize()
        self.setDockNestingEnabled(tprefs['nestable_dock_widgets'])
        for v, h in product(('top', 'bottom'), ('left', 'right')):
            p = f'dock_{v}_{h}'
            pref = tprefs[p] or tprefs.defaults[p]
            area = getattr(Qt.DockWidgetArea, '{}DockWidgetArea'.format(capitalize({'vertical': h, 'horizontal': v}[pref])))
            self.setCorner(getattr(Qt.Corner, '{}{}Corner'.format(*tuple(map(capitalize, (v, h))))), area)
        self.preview.apply_settings()
        self.live_css.apply_theme()
        for bar in (self.global_bar, self.tools_bar, self.plugins_bar):
            bar.setIconSize(QSize(tprefs['toolbar_icon_size'], tprefs['toolbar_icon_size']))

    def show_status_message(self, msg, timeout=5):
        status_bar = self.status_bar
        assert status_bar is not None
        status_bar.showMessage(msg, int(timeout * 1000))

    def update_status_bar_default_message(self, path=''):
        m = self.status_bar_default_msg
        if path:
            m.setText(getattr(m, 'editing_template').format(path=path))
        else:
            m.setText(getattr(m, 'base_template'))

    def elided_text(self, text, width=300):
        return elided_text(text, font=self.font(), width=width)

    @property
    def editor_tabs(self):
        return self.central.editor_tabs

    def create_actions(self):
        group = _('Global actions')

        def reg(icon, text, target, sid, keys, description, toolbar_allowed=False):
            if not isinstance(icon, QIcon):
                icon = QIcon.ic(icon)
            ac = actions[sid] = QAction(icon, text, self) if icon else QAction(text, self)
            ac.setObjectName('action-' + sid)
            if toolbar_allowed:
                toolbar_actions[sid] = ac
            if target is not None:
                ac.triggered.connect(target)
            if isinstance(keys, str):
                keys = (keys,)
            self.keyboard.register_shortcut(sid, str(ac.text()).replace('&', ''), default_keys=keys, description=description, action=ac, group=group)
            self.addAction(ac)
            return ac

        def treg(icon, text, target, sid, keys, description):
            return reg(icon, text, target, sid, keys, description, toolbar_allowed=icon is not None)

        self.action_new_file = treg(
            'document-new.png',
            _('&New file (images/fonts/HTML/etc.)'),
            self.boss.add_file,
            'new-file',
            (),
            _('Create a new file in the current book'),
        )
        self.action_import_files = treg(
            'document-import.png',
            _('&Import files into book'),
            self.boss.add_files,
            'new-files',
            (),
            _('Import files into book'),
        )
        self.action_open_book = treg('document_open.png', _('&Open book'), self.boss.open_book, 'open-book', 'Ctrl+O', _('Open a book'))
        self.action_open_book_folder = treg(
            'mimetypes/dir.png',
            _('Open &folder (unzipped EPUB) as book'),
            partial(self.boss.open_book, open_folder=True),
            'open-folder-as-book',
            (),
            _('Open a folder (unzipped EPUB) as a book'),
        )
        self.action_edit_next_file = treg(
            'arrow-down.png',
            _('Edit &next file'),
            partial(self.boss.edit_next_file, backwards=False),
            'edit-next-file',
            'Ctrl+Alt+Down',
            _('Edit the next file in the spine'),
        )
        self.action_edit_previous_file = treg(
            'arrow-up.png',
            _('Edit &previous file'),
            partial(self.boss.edit_next_file, backwards=True),
            'edit-previous-file',
            'Ctrl+Alt+Up',
            _('Edit the previous file in the spine'),
        )
        # Qt does not generate shortcut overrides for cmd+arrow on os x which
        # means these shortcuts interfere with editing
        self.action_global_undo = treg(
            'back.png',
            _('&Revert to before'),
            self.boss.do_global_undo,
            'global-undo',
            () if ismacos else 'Ctrl+Left',
            _('Revert book to before the last action (Undo)'),
        )
        self.action_global_redo = treg(
            'forward.png',
            _('&Revert to after'),
            self.boss.do_global_redo,
            'global-redo',
            () if ismacos else 'Ctrl+Right',
            _('Revert book state to after the next action (Redo)'),
        )
        self.action_save = treg('save.png', _('&Save'), self.boss.save_book, 'save-book', 'Ctrl+S', _('Save book'))
        self.action_save.setEnabled(False)
        self.action_save_copy = treg('save.png', _('Save a &copy'), self.boss.save_copy, 'save-copy', 'Ctrl+Alt+S', _('Save a copy of the book'))
        self.action_save_copy_edit = treg(
            'save.png',
            _('Save a &copy and edit in new window'),
            partial(self.boss._save_copy, 'edit'),
            'save-copy-edit',
            'Ctrl+Shift+S',
            _('Save a copy of the book and edit it in a new window'),
        )
        self.action_save_copy_replace = treg(
            'save.png',
            _('Save a &copy and edit here'),
            partial(self.boss._save_copy, 'replace'),
            'save-copy-replace',
            'Ctrl+Alt+Shift+S',
            _('Save a copy of the book and edit it in this window'),
        )
        self.action_quit = treg('window-close.png', _('&Quit'), self.boss.quit, 'quit', 'Ctrl+Q', _('Quit'))
        self.action_preferences = treg('config.png', _('&Preferences'), self.boss.preferences, 'preferences', 'Ctrl+P', _('Preferences'))
        self.action_new_book = treg('plus.png', _('Create new, &empty book'), self.boss.new_book, 'new-book', (), _('Create a new, empty book'))
        self.action_import_book = treg(
            'add_book.png',
            _('&Import an HTML or DOCX file as a new book'),
            self.boss.import_book,
            'import-book',
            (),
            _('Import an HTML or DOCX file as a new book'),
        )
        self.action_quick_edit = treg(
            'modified.png',
            _('&Quick open a file to edit'),
            self.boss.quick_open,
            'quick-open',
            ('Ctrl+T'),
            _('Quickly open a file from the book to edit it'),
        )
        self.action_editor_toggle_wrap = treg(
            'format-justify-fill.png',
            _('Toggle code line &wrapping'),
            self.boss.toggle_line_wrapping_in_all_editors,
            'editor-toggle-wrap',
            (),
            _('Toggle line wrapping in all code editor tabs'),
        )

        # Editor actions
        group = _('Editor actions')
        self.action_editor_undo = reg('edit-undo.png', _('&Undo'), self.boss.do_editor_undo, 'editor-undo', 'Ctrl+Z', _('Undo typing'))
        self.action_editor_redo = reg('edit-redo.png', _('R&edo'), self.boss.do_editor_redo, 'editor-redo', 'Ctrl+Y', _('Redo typing'))
        self.action_editor_cut = reg(
            'edit-cut.png',
            _('Cut &text'),
            self.boss.do_editor_cut,
            'editor-cut',
            (
                'Ctrl+X',
                'Shift+Delete',
            ),
            _('Cut text'),
        )
        self.action_editor_copy = reg(
            'edit-copy.png',
            _('&Copy to clipboard'),
            self.boss.do_editor_copy,
            'editor-copy',
            ('Ctrl+C', 'Ctrl+Insert'),
            _('Copy to clipboard'),
        )
        self.action_editor_paste = reg(
            'edit-paste.png',
            _('P&aste from clipboard'),
            self.boss.do_editor_paste,
            'editor-paste',
            (
                'Ctrl+V',
                'Shift+Insert',
            ),
            _('Paste from clipboard'),
        )
        self.action_editor_cut.setEnabled(False)
        self.action_editor_copy.setEnabled(False)
        self.action_editor_undo.setEnabled(False)
        self.action_editor_redo.setEnabled(False)

        # Tool actions
        group = _('Tools')
        self.action_toc = treg('toc.png', _('&Edit Table of Contents'), self.boss.edit_toc, 'edit-toc', (), _('Edit Table of Contents'))
        self.action_inline_toc = treg(
            'chapters.png',
            _('&Insert inline Table of Contents'),
            self.boss.insert_inline_toc,
            'insert-inline-toc',
            (),
            _('Insert inline Table of Contents'),
        )
        self.action_fix_html_current = reg(
            'html-fix.png',
            _('&Fix HTML'),
            partial(self.boss.fix_html, True),
            'fix-html-current',
            (),
            _('Fix HTML in the current file'),
        )
        self.action_fix_html_all = treg(
            'html-fix.png',
            _('&Fix HTML - all files'),
            partial(self.boss.fix_html, False),
            'fix-html-all',
            (),
            _('Fix HTML in all files'),
        )
        self.action_pretty_current = reg(
            'beautify.png',
            _('&Beautify current file'),
            partial(self.boss.pretty_print, True),
            'pretty-current',
            (),
            _('Beautify current file'),
        )
        self.action_pretty_all = treg(
            'beautify.png',
            _('&Beautify all files'),
            partial(self.boss.pretty_print, False),
            'pretty-all',
            (),
            _('Beautify all files'),
        )
        self.action_insert_char = treg(
            'character-set.png',
            _('&Insert special character'),
            self.boss.insert_character,
            'insert-character',
            (),
            _('Insert special character'),
        )
        self.action_rationalize_folders = treg(
            'mimetypes/dir.png',
            _('&Arrange into folders'),
            self.boss.rationalize_folders,
            'rationalize-folders',
            (),
            _('Arrange into folders'),
        )
        self.action_set_semantics = treg('tags.png', _('Set &semantics'), self.boss.set_semantics, 'set-semantics', (), _('Set semantics'))
        self.action_filter_css = treg(
            'filter.png',
            _('&Filter style information'),
            self.boss.filter_css,
            'filter-css',
            (),
            _('Filter style information'),
        )
        self.action_manage_fonts = treg('font.png', _('&Manage fonts'), self.boss.manage_fonts, 'manage-fonts', (), _('Manage fonts in the book'))
        self.action_add_cover = treg('default_cover.png', _('Add &cover'), self.boss.add_cover, 'add-cover', (), _('Add a cover to the book'))
        self.action_reports = treg(
            'reports.png',
            _('&Reports'),
            self.boss.show_reports,
            'show-reports',
            ('Ctrl+Shift+R',),
            _('Show a report on various aspects of the book'),
        )
        self.action_check_external_links = treg(
            'insert-link.png',
            _('Check &external links'),
            self.boss.check_external_links,
            'check-external-links',
            (),
            _('Check external links in the book'),
        )
        self.action_compress_images = treg(
            'compress-image.png',
            _('C&ompress images losslessly'),
            self.boss.compress_images,
            'compress-images',
            (),
            _('Compress images losslessly'),
        )
        self.action_transform_styles = treg(
            'wizard.png',
            _('Transform &styles'),
            self.boss.transform_styles,
            'transform-styles',
            (),
            _('Transform styles used in the book'),
        )
        self.action_transform_html = treg(
            'wizard.png',
            _('Transform &HTML'),
            self.boss.transform_html,
            'transform-html',
            (),
            _('Transform HTML used in the book'),
        )
        self.action_get_ext_resources = treg(
            'download-metadata.png',
            _('Download external &resources'),
            self.boss.get_external_resources,
            'get-external-resources',
            (),
            _('Download external resources in the book (images/stylesheets/etc/ that are not included in the book)'),
        )
        self.action_embed_tts = treg(
            'bullhorn.png',
            _('Add Text-to-speech narration'),
            self.boss.embed_tts,
            'embed-tts',
            (),
            _('Add audio narration for all the book text using Text-to-speech generation'),
        )

        def ereg(icon, text, target, sid, keys, description):
            return reg(icon, text, partial(self.boss.editor_action, target), sid, keys, description)

        register_text_editor_actions(ereg, self.palette())

        # Polish actions
        group = _('Polish book')
        self.action_subset_fonts = treg(
            'subset-fonts.png',
            _('&Subset embedded fonts'),
            partial(self.boss.polish, 'subset', _('Subset fonts')),
            'subset-fonts',
            (),
            _('Subset embedded fonts'),
        )
        self.action_embed_fonts = treg(
            'embed-fonts.png',
            _('&Embed referenced fonts'),
            partial(self.boss.polish, 'embed', _('Embed fonts')),
            'embed-fonts',
            (),
            _('Embed referenced fonts'),
        )
        self.action_smarten_punctuation = treg(
            'smarten-punctuation.png',
            _('&Smarten punctuation (works best for English)'),
            partial(self.boss.polish, 'smarten_punctuation', _('Smarten punctuation')),
            'smarten-punctuation',
            (),
            _('Smarten punctuation'),
        )
        self.action_remove_unused_css = treg(
            'edit-clear.png',
            _('Remove &unused CSS rules'),
            partial(self.boss.polish, 'remove_unused_css', _('Remove unused CSS rules')),
            'remove-unused-css',
            (),
            _('Remove unused CSS rules'),
        )
        self.action_remove_unused_images = treg(
            'edit-clear.png',
            _('Remove unused &images'),
            partial(self.boss.polish, 'remove_unused_images', _('Remove unused images')),
            'remove-unused-images',
            (),
            _('Remove unused images'),
        )
        self.action_upgrade_book_internals = treg(
            'arrow-up.png',
            _('&Upgrade book internals'),
            partial(self.boss.polish, 'upgrade_book', _('Upgrade book internals')),
            'upgrade-book',
            (),
            _('Upgrade book internals'),
        )

        # Preview actions
        group = _('Preview')
        self.action_auto_reload_preview = reg('auto-reload.png', _('Auto reload preview'), None, 'auto-reload-preview', (), _('Auto reload preview'))
        self.action_auto_sync_preview = reg(
            'sync-right.png',
            _('Sync preview position to editor position'),
            None,
            'sync-preview-to-editor',
            (),
            _('Sync preview position to editor position'),
        )
        self.action_reload_preview = reg('view-refresh.png', _('Refresh preview'), None, 'reload-preview', ('F5',), _('Refresh preview'))
        self.action_split_in_preview = reg(
            'document-split.png',
            _('Split this file'),
            None,
            'split-in-preview',
            (),
            _('Split file in the preview panel'),
        )
        self.action_find_next_preview = reg('arrow-down.png', _('Find next'), None, 'find-next-preview', (), _('Find next in preview'))
        self.action_find_prev_preview = reg('arrow-up.png', _('Find previous'), None, 'find-prev-preview', (), _('Find previous in preview'))
        self.action_copy_from_preview = reg(
            'edit-copy.png',
            _('Copy from preview'),
            None,
            'copy-from-preview',
            ('Ctrl+Alt+c',),
            _('Copy the current selection from the preview panel to the clipboard'),
        )

        # Search actions
        group = _('Search')
        self.action_find = treg(
            'search.png',
            _('&Find/replace'),
            self.boss.show_find,
            'find-replace',
            ('Ctrl+F',),
            _('Show the Find/replace panel'),
        )

        def sreg(name, text, action, overrides={}, keys=(), description=None, icon=None):
            return reg(
                icon,
                text,
                partial(self.boss.search_action_triggered, action, overrides),
                name,
                keys,
                description or text.replace('&', ''),
            )

        self.action_find_next = sreg('find-next', _('Find &next'), 'find', {'direction': 'down'}, ('F3', 'Ctrl+G'), _('Find next match'))
        self.action_find_previous = sreg(
            'find-previous',
            _('Find &previous'),
            'find',
            {'direction': 'up'},
            ('Shift+F3', 'Shift+Ctrl+G'),
            _('Find previous match'),
        )
        self.action_replace = sreg('replace', _('&Replace'), 'replace', keys=('Ctrl+R'), description=_('Replace current match'))
        self.action_replace_next = sreg(
            'replace-next',
            _('Replace and find ne&xt'),
            'replace-find',
            {'direction': 'down'},
            ('Ctrl+]'),
            _('Replace current match and find next'),
        )
        self.action_replace_previous = sreg(
            'replace-previous',
            _('R&eplace and find previous'),
            'replace-find',
            {'direction': 'up'},
            ('Ctrl+['),
            _('Replace current match and find previous'),
        )
        self.action_replace_all = sreg('replace-all', _('Replace &all'), 'replace-all', keys=('Ctrl+A'), description=_('Replace all matches'))
        self.action_count = sreg('count-matches', _('&Count all'), 'count', keys=('Ctrl+N'), description=_('Count number of matches'))
        self.action_mark = reg(
            None,
            _('&Mark selected text'),
            self.boss.mark_selected_text,
            'mark-selected-text',
            ('Ctrl+Shift+M',),
            _('Mark selected text or unmark already marked text'),
        )
        self.action_mark.default_text = self.action_mark.text()
        self.action_go_to_line = reg(
            None,
            _('Go to &line'),
            self.boss.go_to_line_number,
            'go-to-line-number',
            ('Ctrl+.',),
            _('Go to line number'),
        )
        self.action_saved_searches = treg(
            'folder_saved_search.png',
            _('Sa&ved searches'),
            self.boss.saved_searches,
            'saved-searches',
            (),
            _('Show the saved searches dialog'),
        )
        self.action_text_search = treg(
            'view.png',
            _('&Search ignoring HTML markup'),
            self.boss.show_text_search,
            'text-search',
            (),
            _('Show the text search panel'),
        )

        # Check Book actions
        group = _('Check book')
        self.action_check_book = treg('debug.png', _('C&heck book'), self.boss.check_requested, 'check-book', ('F7'), _('Check book for errors'))
        self.action_spell_check_book = treg(
            'spell-check.png',
            _('Check &spelling'),
            self.boss.spell_check_requested,
            'spell-check-book',
            ('Alt+F7'),
            _('Check book for spelling errors'),
        )
        self.action_check_book_next = reg(
            'forward.png',
            _('&Next error'),
            partial(self.check_book.next_error, delta=1),
            'check-book-next',
            ('Ctrl+F7'),
            _('Show next error'),
        )
        self.action_check_book_previous = reg(
            'back.png',
            _('&Previous error'),
            partial(self.check_book.next_error, delta=-1),
            'check-book-previous',
            ('Ctrl+Shift+F7'),
            _('Show previous error'),
        )
        self.action_spell_check_next = reg(
            'forward.png',
            _('&Next spelling mistake'),
            self.boss.next_spell_error,
            'spell-next',
            ('F8'),
            _('Go to next spelling mistake'),
        )

        # Miscellaneous actions
        group = pgettext('edit book actions', 'Miscellaneous')
        self.action_create_checkpoint = treg(
            'marked.png',
            _('&Create checkpoint'),
            self.boss.create_checkpoint,
            'create-checkpoint',
            (),
            _('Create a checkpoint with the current state of the book'),
        )
        self.action_close_current_tab = reg(
            'window-close.png',
            _('&Close current tab'),
            self.central.close_current_editor,
            'close-current-tab',
            'Ctrl+W',
            _('Close the currently open tab'),
        )
        self.action_close_all_but_current_tab = reg(
            'edit-clear.png',
            _('C&lose other tabs'),
            self.central.close_all_but_current_editor,
            'close-all-but-current-tab',
            'Ctrl+Alt+W',
            _('Close all tabs except the current tab'),
        )
        self.action_close_to_right = reg(
            'edit-clear.png',
            _('Close tabs to the &right'),
            self.central.close_to_right_of_current_editor,
            'close-tabs-to-right-of',
            'Ctrl+Shift+W',
            _('Close tabs to the right of the current tab'),
        )
        self.action_help = treg(
            'help.png',
            _('User &Manual'),
            lambda: open_url(QUrl(localize_user_manual_link('https://manual.calibre-ebook.com/edit.html'))),
            'user-manual',
            'F1',
            _('Show User Manual'),
        )
        self.action_browse_images = treg(
            'view-image.png',
            _('&Browse images in book'),
            self.boss.browse_images,
            'browse-images',
            (),
            _('Browse images in the books visually'),
        )
        self.action_multiple_split = treg(
            'document-split.png',
            _('&Split at multiple locations'),
            self.boss.multisplit,
            'multisplit',
            (),
            _('Split HTML file at multiple locations'),
        )
        self.action_compare_book = treg(
            'diff.png',
            _('Compare to &another book'),
            self.boss.compare_book,
            'compare-book',
            (),
            _('Compare to another book'),
        )
        self.action_manage_snippets = treg(
            'snippets.png',
            _('Manage &Snippets'),
            self.boss.manage_snippets,
            'manage-snippets',
            (),
            _('Manage user created Snippets'),
        )
        self.action_merge_files = treg(
            'merge.png',
            _('&Merge files'),
            self.boss.merge_files,
            'merge-files',
            'Ctrl+M',
            _('Merge two or more selected files'),
        )

        self.plugin_menu_actions = []

        create_plugin_actions(actions, toolbar_actions, self.plugin_menu_actions)

    def create_menubar(self):
        if ismacos:
            p, q = self.create_application_menubar()
            q.triggered.connect(self.action_quit.trigger)
            p.triggered.connect(self.action_preferences.trigger)
        b = QMenuBar(self)
        self.setMenuBar(b)
        setattr(b, 'is_native_menubar', False)

        f = b.addMenu(_('&File'))
        assert f is not None
        f.addAction(self.action_new_file)
        f.addAction(self.action_import_files)
        f.addSeparator()
        f.addAction(self.action_open_book)
        f.addAction(self.action_new_book)
        f.addAction(self.action_import_book)
        f.addAction(self.action_open_book_folder)
        self.recent_books_menu = f.addMenu(_('&Recently opened books'))
        self.update_recent_books()
        f.addSeparator()
        f.addAction(self.action_save)
        m = f.addMenu(_('Save a copy'))
        assert m is not None
        m.addAction(self.action_save_copy)
        m.addAction(self.action_save_copy_edit)
        m.addAction(self.action_save_copy_replace)
        f.addSeparator()
        f.addAction(self.action_compare_book)
        f.addAction(self.action_quit)

        e = b.addMenu(_('&Edit'))
        assert e is not None
        e.addAction(self.action_global_undo)
        e.addAction(self.action_global_redo)
        e.addAction(self.action_create_checkpoint)
        e.addSeparator()
        e.addAction(self.action_editor_undo)
        e.addAction(self.action_editor_redo)
        e.addSeparator()
        e.addAction(self.action_editor_cut)
        e.addAction(self.action_editor_copy)
        e.addAction(self.action_editor_paste)
        e.addAction(self.action_insert_char)
        e.addSeparator()
        e.addAction(self.action_quick_edit)
        e.addAction(self.action_preferences)

        e = b.addMenu(_('&Tools'))
        assert e is not None
        tm = e.addMenu(QIcon.ic('toc.png'), _('Table of Contents'))
        assert tm is not None
        tm.addAction(self.action_toc)
        tm.addAction(self.action_inline_toc)
        e.addAction(self.action_manage_fonts)
        e.addAction(self.action_embed_fonts)
        e.addAction(self.action_subset_fonts)
        e.addAction(self.action_compress_images)
        e.addAction(self.action_smarten_punctuation)
        ru = e.addMenu(QIcon.ic('edit-clear.png'), _('Remove &unused') + '…')
        assert ru is not None
        ru.addAction(self.action_remove_unused_css)
        ru.addAction(self.action_remove_unused_images)
        ts = e.addMenu(QIcon.ic('wizard.png'), _('Transform markup/styles'))
        assert ts is not None
        ts.addAction(self.action_transform_styles)
        ts.addAction(self.action_transform_html)
        e.addAction(self.action_fix_html_all)
        e.addAction(self.action_embed_tts)
        e.addAction(self.action_pretty_all)
        e.addAction(self.action_rationalize_folders)
        e.addAction(self.action_add_cover)
        e.addAction(self.action_set_semantics)
        e.addAction(self.action_filter_css)
        e.addAction(self.action_spell_check_book)
        er = e.addMenu(QIcon.ic('external-link.png'), _('External &links'))
        assert er is not None
        er.addAction(self.action_check_external_links)
        er.addAction(self.action_get_ext_resources)
        e.addAction(self.action_check_book)
        e.addAction(self.action_reports)
        e.addAction(self.action_upgrade_book_internals)

        e = b.addMenu(_('&View'))
        assert e is not None
        t = e.addMenu(_('Tool&bars'))
        assert t is not None
        e.addSeparator()
        for name in sorted(actions, key=lambda x: sort_key(actions[x].text())):
            ac = actions[name]
            if name.endswith('-dock'):
                e.addAction(ac)
            elif name.endswith('-bar'):
                t.addAction(ac)
        e.addAction(self.action_browse_images)
        e.addSeparator()
        e.addAction(self.action_close_current_tab)
        e.addAction(self.action_close_all_but_current_tab)

        e = b.addMenu(_('&Search'))
        assert e is not None
        a = e.addAction
        a(self.action_find)
        e.addSeparator()
        a(self.action_find_next)
        a(self.action_find_previous)
        e.addSeparator()
        a(self.action_replace)
        a(self.action_replace_next)
        a(self.action_replace_previous)
        a(self.action_replace_all)
        e.addSeparator()
        a(self.action_count)
        e.addSeparator()
        a(self.action_mark)
        e.addSeparator()
        a(self.action_go_to_line)
        e.addSeparator()
        a(self.action_saved_searches)
        e.aboutToShow.connect(self.search_menu_about_to_show)
        e.addSeparator()
        a(self.action_text_search)

        if self.plugin_menu_actions:
            e = b.addMenu(_('&Plugins'))
            assert e is not None
            for ac in sorted(self.plugin_menu_actions, key=lambda x: sort_key(str(x.text()))):
                e.addAction(ac)

        e = b.addMenu(_('&Help'))
        assert e is not None
        a = e.addAction
        a(self.action_help)
        a(QIcon.ic('donate.png'), _('&Donate to support calibre development'), open_donate)
        a(self.action_preferences)

    def search_menu_about_to_show(self):
        ed = self.central.current_editor
        update_mark_text_action(ed)

    def update_recent_books(self):
        m = self.recent_books_menu
        assert m is not None
        m.clear()
        books = tprefs.get('recent-books', [])
        for path in books:
            m.addAction(self.elided_text(path, width=500), partial(self.boss.open_book, path=path))

    def create_toolbars(self):
        def create(text, name):
            name += '-bar'
            b = self.addToolBar(text)
            b.setObjectName(name)  # Needed for saveState
            actions[name] = b.toggleViewAction()
            b.setIconSize(QSize(tprefs['toolbar_icon_size'], tprefs['toolbar_icon_size']))
            return b

        self.global_bar = create(_('Book tool bar'), 'global')
        self.tools_bar = create(_('Tools tool bar'), 'tools')
        self.plugins_bar = create(_('Plugins tool bar'), 'plugins')
        self.populate_toolbars(animate=True)

    def populate_toolbars(self, animate=False):
        self.global_bar.clear(), self.tools_bar.clear(), self.plugins_bar.clear()

        def add(bar, ac):
            if ac is None:
                bar.addSeparator()
            elif ac == 'donate':
                b = ThrobbingButton(self)
                b.clicked.connect(open_donate)
                b.setAutoRaise(True)
                b.setToolTip(_('Donate to support calibre development'))
                if animate:
                    QTimer.singleShot(10, b.start_animation)
                bar.addWidget(b)
            else:
                try:
                    bar.addAction(actions[ac])
                except KeyError:
                    if DEBUG:
                        prints(f'Unknown action for toolbar {str(bar.objectName())!r}: {ac!r}')

        for x in tprefs['global_book_toolbar']:
            add(self.global_bar, x)

        for x in tprefs['global_tools_toolbar']:
            add(self.tools_bar, x)

        for x in tprefs['global_plugins_toolbar']:
            add(self.plugins_bar, x)
        self.plugins_bar.setVisible(bool(tprefs['global_plugins_toolbar']))

    def create_docks(self):

        def create(name, oname):
            oname += '-dock'
            d = QDockWidget(name, self)
            d.setObjectName(oname)  # Needed for saveState
            ac = d.toggleViewAction()
            desc = _('Toggle %s') % name.replace('&', '')
            self.keyboard.register_shortcut(oname, desc, description=desc, action=ac, group=_('Windows'))
            actions[oname] = ac
            setattr(self, oname.replace('-', '_'), d)
            return d

        d = create(_('File browser'), 'files-browser')
        d.setAllowedAreas(Qt.DockWidgetArea.LeftDockWidgetArea | Qt.DockWidgetArea.RightDockWidgetArea)
        self.file_list = FileListWidget(d)
        d.setWidget(self.file_list)
        self.addDockWidget(Qt.DockWidgetArea.LeftDockWidgetArea, d)

        d = create(_('File preview'), 'preview')
        d.setAllowedAreas(Qt.DockWidgetArea.LeftDockWidgetArea | Qt.DockWidgetArea.RightDockWidgetArea)
        self.preview = Preview(d)
        d.setWidget(self.preview)
        self.addDockWidget(Qt.DockWidgetArea.RightDockWidgetArea, d)

        d = create(_('Live CSS'), 'live-css')
        d.setAllowedAreas(
            Qt.DockWidgetArea.LeftDockWidgetArea
            | Qt.DockWidgetArea.RightDockWidgetArea
            | Qt.DockWidgetArea.BottomDockWidgetArea
            | Qt.DockWidgetArea.TopDockWidgetArea
        )
        self.live_css = LiveCSS(self.preview, parent=d)
        d.setWidget(self.live_css)
        self.addDockWidget(Qt.DockWidgetArea.RightDockWidgetArea, d)
        d.close()  # Hidden by default

        d = create(_('Check book'), 'check-book')
        d.setAllowedAreas(
            Qt.DockWidgetArea.LeftDockWidgetArea
            | Qt.DockWidgetArea.RightDockWidgetArea
            | Qt.DockWidgetArea.BottomDockWidgetArea
            | Qt.DockWidgetArea.TopDockWidgetArea
        )
        d.setWidget(self.check_book)
        self.addDockWidget(Qt.DockWidgetArea.TopDockWidgetArea, d)
        d.close()  # By default the check window is closed

        d = create(_('Inspector'), 'inspector')
        d.setAllowedAreas(
            Qt.DockWidgetArea.LeftDockWidgetArea
            | Qt.DockWidgetArea.RightDockWidgetArea
            | Qt.DockWidgetArea.BottomDockWidgetArea
            | Qt.DockWidgetArea.TopDockWidgetArea
        )
        d.setWidget(self.preview.inspector)
        self.preview.inspector.setParent(d)
        self.addDockWidget(Qt.DockWidgetArea.BottomDockWidgetArea, d)
        d.close()  # By default the inspector window is closed
        QTimer.singleShot(10, self.preview.inspector.connect_to_dock)

        d = create(_('Table of Contents'), 'toc-viewer')
        d.setAllowedAreas(
            Qt.DockWidgetArea.LeftDockWidgetArea
            | Qt.DockWidgetArea.RightDockWidgetArea
            | Qt.DockWidgetArea.BottomDockWidgetArea
            | Qt.DockWidgetArea.TopDockWidgetArea
        )
        d.setWidget(self.toc_view)
        self.addDockWidget(Qt.DockWidgetArea.LeftDockWidgetArea, d)
        d.close()  # Hidden by default

        d = create(_('Text search'), 'text-search')
        d.setAllowedAreas(
            Qt.DockWidgetArea.LeftDockWidgetArea
            | Qt.DockWidgetArea.RightDockWidgetArea
            | Qt.DockWidgetArea.BottomDockWidgetArea
            | Qt.DockWidgetArea.TopDockWidgetArea
        )
        d.setWidget(self.text_search)
        self.addDockWidget(Qt.DockWidgetArea.LeftDockWidgetArea, d)
        d.close()  # Hidden by default

        d = create(_('Checkpoints'), 'checkpoints')
        d.setAllowedAreas(
            Qt.DockWidgetArea.LeftDockWidgetArea
            | Qt.DockWidgetArea.RightDockWidgetArea
            | Qt.DockWidgetArea.BottomDockWidgetArea
            | Qt.DockWidgetArea.TopDockWidgetArea
        )
        self.checkpoints = CheckpointView(self.boss.global_undo, parent=d)
        d.setWidget(self.checkpoints)
        self.addDockWidget(Qt.DockWidgetArea.LeftDockWidgetArea, d)
        d.close()  # Hidden by default

        d = create(_('Saved searches'), 'saved-searches')
        d.setAllowedAreas(
            Qt.DockWidgetArea.LeftDockWidgetArea
            | Qt.DockWidgetArea.RightDockWidgetArea
            | Qt.DockWidgetArea.BottomDockWidgetArea
            | Qt.DockWidgetArea.TopDockWidgetArea
        )
        d.setWidget(self.saved_searches)
        self.addDockWidget(Qt.DockWidgetArea.LeftDockWidgetArea, d)
        d.close()  # Hidden by default

    def resizeEvent(self, a0):
        self.blocking_job.resize(a0.size())
        return super().resizeEvent(a0)

    def update_window_title(self):
        cc = current_container()
        if cc is not None:
            fname = os.path.basename(cc.path_to_ebook)
            current_metadata = self.current_metadata
            assert current_metadata is not None
            self.setWindowTitle(current_metadata.title + f' [{cc.book_type_for_display}] :: {fname} :: {self.APP_NAME}')
        else:
            self.setWindowTitle(self.APP_NAME)

    def closeEvent(self, a0):
        if self.boss.quit():
            a0.accept()
        else:
            a0.ignore()

    def save_state(self):
        self.save_geometry(tprefs, 'main_window_geometry')
        tprefs.set('main_window_state', bytearray(self.saveState(self.STATE_VERSION)))
        self.central.save_state()
        self.saved_searches.save_state()
        self.check_book.save_state()
        self.text_search.save_state()

    def restore_state(self):
        self.restore_geometry(tprefs, 'main_window_geometry')
        state = tprefs.get('main_window_state', None)
        if state is not None:
            self.restoreState(state, self.STATE_VERSION)
        self.central.restore_state()
        self.saved_searches.restore_state()

    def contextMenuEvent(self, event):
        event.ignore()



#!/usr/bin/env python
# License: GPLv3 Copyright: 2014, Kovid Goyal <kovid at kovidgoyal.net>

from collections import defaultdict, namedtuple
from operator import itemgetter

from lxml import etree
from qt.core import QCheckBox, QDialog, QDialogButtonBox, QFormLayout, QHBoxLayout, QIcon, QLabel, QLineEdit, QSpinBox, Qt, QToolButton

from calibre.gui2 import choose_files, error_dialog
from calibre.utils.icu import sort_key
from calibre.utils.localization import _
from calibre.utils.xml_parse import safe_xml_fromstring

Group = namedtuple('Group', 'title feeds')


def uniq(vals, kmap=lambda x: x):
    """Remove all duplicates from vals, while preserving order. kmap must be a
    callable that returns a hashable value for every item in vals"""
    vals = vals or ()
    lvals = (kmap(x) for x in vals)
    seen = set()
    seen_add = seen.add
    return tuple(x for x, k in zip(vals, lvals) if k not in seen and not seen_add(k))


def import_opml(raw, preserve_groups=True):
    root = safe_xml_fromstring(raw)
    groups = defaultdict(list)
    ax = etree.XPath('ancestor::outline[@title or @text]')
    for outline in root.xpath('//outline[@type="rss" and @xmlUrl]'):
        url = outline.get('xmlUrl')
        parent = outline.get('title', '') or url
        title = parent if ('title' in outline.attrib and parent) else None
        if preserve_groups:
            for ancestor in ax(outline):
                if ancestor.get('type', None) != 'rss':
                    text = ancestor.get('title') or ancestor.get('text')
                    if text:
                        parent = text
                        break
        groups[parent].append((title, url))

    for title in sorted(groups, key=sort_key):
        yield Group(title, uniq(groups[title], kmap=itemgetter(1)))


class ImportOPML(QDialog):
    def __init__(self, parent=None):
        QDialog.__init__(self, parent=parent)
        self.l = l = QFormLayout(self)
        self.setLayout(l)
        self.setWindowTitle(_('Import OPML file'))
        self.setWindowIcon(QIcon.ic('mimetypes/opml.png'))

        self.h = h = QHBoxLayout()
        self.path = p = QLineEdit(self)
        p.setMinimumWidth(300)
        p.setPlaceholderText(_('Path to OPML file'))
        h.addWidget(p)
        self.cfb = b = QToolButton(self)
        b.setIcon(QIcon.ic('document_open.png'))
        b.setToolTip(_('Browse for OPML file'))
        b.clicked.connect(self.choose_file)
        h.addWidget(b)
        l.addRow(_('&OPML file:'), h)
        lbl = l.labelForField(h)
        assert isinstance(lbl, QLabel)
        lbl.setBuddy(p)
        b.setFocus(Qt.FocusReason.OtherFocusReason)

        self._articles_per_feed = a = QSpinBox(self)
        a.setMinimum(1), a.setMaximum(1000), a.setValue(100)
        a.setToolTip(_('Maximum number of articles to download per RSS feed'))
        l.addRow(_('&Maximum articles per feed:'), a)

        self._oldest_article = o = QSpinBox(self)
        o.setMinimum(1), o.setMaximum(3650), o.setValue(7)
        o.setSuffix(_(' days'))
        o.setToolTip(_('Articles in the RSS feeds older than this will be ignored'))
        l.addRow(_('&Oldest article:'), o)

        self.preserve_groups = g = QCheckBox(_('Preserve groups in the OPML file'))
        g.setToolTip(
            '<p>' + _('If enabled, every group of feeds in the OPML file will be converted into a single recipe. Otherwise every feed becomes its own recipe')
        )
        g.setChecked(True)
        l.addRow(g)

        self._replace_existing = r = QCheckBox(_('Replace existing recipes'))
        r.setToolTip(
            '<p>'
            + _(
                'If enabled, any existing recipes with the same titles as entries in the OPML file will be replaced.'
                ' Otherwise, new entries with modified titles will be created'
            )
        )
        r.setChecked(True)
        l.addRow(r)

        self.bb = bb = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        bb.accepted.connect(self.accept), bb.rejected.connect(self.reject)
        l.addRow(bb)

        self.recipes = ()

    @property
    def articles_per_feed(self):
        return self._articles_per_feed.value()

    @property
    def oldest_article(self):
        return self._oldest_article.value()

    @property
    def replace_existing(self):
        return self._replace_existing.isChecked()

    def choose_file(self):
        opml_files = choose_files(
            self,
            'opml-select-dialog',
            _('Select OPML file'),
            filters=[(_('OPML files'), ['opml'])],
            all_files=False,
            select_only_single_file=True,
        )
        if opml_files:
            self.path.setText(opml_files[0])

    def accept(self):
        path = str(self.path.text())
        if not path:
            return error_dialog(self, _('Path not specified'), _('You must specify the path to the OPML file to import'), show=True)
        with open(path, 'rb') as f:
            raw = f.read()
        self.recipes = tuple(import_opml(raw, self.preserve_groups.isChecked()))
        if len(self.recipes) == 0:
            return error_dialog(self, _('No feeds found'), _('No importable RSS feeds found in the OPML file'), show=True)

        QDialog.accept(self)


if __name__ == '__main__':
    import sys

    for group in import_opml(open(sys.argv[-1], 'rb').read()):
        print(group.title)
        for title, url in group.feeds:
            print(f'\t{title} - {url}')
        print()



#!/usr/bin/env python
# License: GPLv3 Copyright: 2013, Kovid Goyal <kovid at kovidgoyal.net>

from collections.abc import Generator
from contextlib import contextmanager

from calibre.ebooks import DRMError as _DRMError
from calibre.utils.localization import _


class InvalidBook(ValueError):
    pass


_drm_message = ''


@contextmanager
def drm_message(msg: str) -> Generator[None]:
    global _drm_message
    orig, _drm_message = _drm_message, msg
    try:
        yield
    finally:
        _drm_message = orig


class DRMError(_DRMError):
    def __init__(self):
        super().__init__(_drm_message or _('This file is locked with DRM. It cannot be edited.'))


class MalformedMarkup(ValueError):
    pass


class UnsupportedContainerType(Exception):
    pass



# License: GPLv3 Copyright: 2009, John Schember <john at nachtimwald.com>

"""
Device driver for Barns and Nobel's Nook
"""

import io
import os

from calibre import fsync, prints
from calibre.constants import DEBUG
from calibre.devices.usbms.driver import USBMS
from calibre.utils.localization import _
from calibre.utils.resources import get_image_path as I


class NOOK(USBMS):
    name = 'Nook Device Interface'
    gui_name = 'B&N Nook'
    description = _('Communicate with the Nook e-book reader.')
    author = 'John Schember'
    icon = 'devices/nook.png'
    supported_platforms = ['windows', 'linux', 'osx']

    # Ordered list of supported formats
    FORMATS = ['epub', 'pdb', 'pdf']

    VENDOR_ID = [0x2080, 0x18D1]  # 0x18d1 is for softrooted nook
    PRODUCT_ID = [0x001]
    BCD = [0x322]

    VENDOR_NAME = 'B&N'
    WINDOWS_MAIN_MEM = 'NOOK'
    WINDOWS_CARD_A_MEM = 'NOOK'

    OSX_MAIN_MEM = 'B&N nook Media'
    OSX_CARD_A_MEM = OSX_MAIN_MEM

    MAIN_MEMORY_VOLUME_LABEL = 'Nook Main Memory'
    STORAGE_CARD_VOLUME_LABEL = 'Nook Storage Card'

    EBOOK_DIR_MAIN = 'my documents'
    THUMBNAIL_HEIGHT = 144
    DELETE_EXTS = ['.jpg']
    SUPPORTS_SUB_DIRS = True

    def upload_cover(self, path, filename, metadata, filepath):
        from PIL import Image, ImageDraw

        coverdata = getattr(metadata, 'thumbnail', None)
        if coverdata and coverdata[2]:
            cover = Image.open(io.BytesIO(coverdata[2]))
        else:
            coverdata = open(I('library.png'), 'rb').read()

            cover = Image.new('RGB', (96, 144), 'black')
            im = Image.open(io.BytesIO(coverdata))
            im.thumbnail((96, 144), Image.Resampling.LANCZOS)

            x, y = im.size
            cover.paste(im, ((96 - x) // 2, (144 - y) // 2))

            draw = ImageDraw.Draw(cover)
            draw.text((1, 15), metadata.get('title', _('Unknown')).encode('ascii', 'ignore'))
            draw.text((1, 115), metadata.get('authors', _('Unknown')).encode('ascii', 'ignore'))

        data = io.BytesIO()
        cover.save(data, 'JPEG')
        coverdata = data.getvalue()

        with open(f'{os.path.join(path, filename)}.jpg', 'wb') as coverfile:
            coverfile.write(coverdata)
            fsync(coverfile)

    def sanitize_path_components(self, components):
        return [x.replace('#', '_').replace('%', '_') for x in components]


class NOOK_COLOR(NOOK):
    name = 'Nook Color Device Interface'
    gui_name = _('B&N Nook Color')
    description = _('Communicate with the Nook Color, TSR, Glowlight and Tablet e-book readers.')

    PRODUCT_ID = [
        0x002,
        0x003,
        0x004,
        0x005,  # Nook HD+
        0x007,  # Glowlight from 2013
        # 0xa,    # Glowlight from 2016 is MTP based device
        0xB,  # Glowlight from 2017
        0xC,  # Glowlight from 2019
        0xD,  # Glowlight from 2021
        0xE,  # Glowlight from 2022
        # 0xf,    # Glowlight from 2023 is MTP based device
    ]
    BCD = [0x216, 0x9999, 0x409, 0x440]

    WINDOWS_MAIN_MEM = WINDOWS_CARD_A_MEM = ['EBOOK_DISK', 'NOOK_TABLET', 'NOOK_SIMPLETOUCH', 'NOOK_GLOWLIGHT']
    EBOOK_DIR_MAIN = 'My Files'
    SCAN_FROM_ROOT = True
    NEWS_IN_FOLDER = False

    def upload_cover(self, path, filename, metadata, filepath):
        pass

    def post_open_callback(self):
        device_info = self.device_being_opened
        if device_info is None:
            return
        product_id = device_info[1]
        if DEBUG:
            prints('Opened NOOK with product id:', product_id)
        if product_id >= 0xB:
            self.EBOOK_DIR_MAIN = 'NOOK/Books' if product_id >= 0xD else 'NOOK/My Files'
            if DEBUG:
                prints(f'Setting Nook upload directory to {self.EBOOK_DIR_MAIN}')
            try:
                assert self._main_prefix is not None
                os.makedirs(os.path.join(self._main_prefix, *self.EBOOK_DIR_MAIN.split('/')), exist_ok=True)
            except OSError:
                self.EBOOK_DIR_MAIN = 'NOOK'

    def get_carda_ebook_dir(self, for_upload=False):
        if for_upload:
            return self.EBOOK_DIR_MAIN
        return ''

    def create_upload_path(self, path, mdata, fname, create_dirs=True):
        is_news = mdata.tags and _('News') in mdata.tags
        subdir = 'Magazines' if is_news else 'Books'
        path = os.path.join(path, subdir)
        return USBMS.create_upload_path(self, path, mdata, fname, create_dirs=create_dirs)



# License: GPLv3 Copyright: 2010, John Schember <john@nachtimwald.com>

"""
Read content from palmdoc pdb file.
"""

from calibre.ebooks.pdb.formatreader import FormatReader
from calibre.ptempfile import PersistentTemporaryFile


class Reader(FormatReader):
    def __init__(self, header, stream, log, options):
        self.header = header
        self.stream = stream
        self.log = log
        self.options = options

    def extract_content(self, output_dir):
        self.log.info('Extracting PDF...')

        pdf = PersistentTemporaryFile('.pdf')
        pdf_path = pdf.name
        pdf.close()
        with open(pdf_path, 'wb') as pdf_out:
            for x in range(self.header.section_count()):
                pdf_out.write(self.header.section_data(x))

        from calibre.customize.ui import plugin_for_input_format

        pdf_plugin = plugin_for_input_format('pdf')
        for opt in pdf_plugin.options:
            if not hasattr(self.options, opt.option.name):
                setattr(self.options, opt.option.name, opt.recommended_value)

        return pdf_plugin.convert(open(pdf_path, 'rb'), self.options, 'pdf', self.log, {})



#!/usr/bin/env python
# License: GPLv3 Copyright: 2011, Kovid Goyal <kovid@kovidgoyal.net>

import numbers
import textwrap

from qt.core import QCheckBox, QComboBox, QDoubleSpinBox, QGridLayout, QGroupBox, QLabel, QLineEdit, QListView, QSpinBox, Qt, QVBoxLayout, QWidget

from calibre.gui2.preferences.metadata_sources import FieldsModel as FM
from calibre.utils.icu import sort_key
from calibre.utils.localization import _


class FieldsModel(FM):  # {{{
    def __init__(self, plugin):
        FM.__init__(self)
        self.plugin = plugin
        self.exclude = frozenset(['title', 'authors']) | self.exclude
        self.prefs = self.plugin.prefs

    def initialize(self):
        fields = self.plugin.touched_fields
        self.beginResetModel()
        self.fields = []
        for x in fields:
            if not x.startswith('identifier:') and x not in self.exclude:
                self.fields.append(x)
        self.fields.sort(key=lambda x: self.descs.get(x, x))
        self.endResetModel()

    def state(self, field, defaults=False):
        src = self.prefs.defaults if defaults else self.prefs
        return Qt.CheckState.Unchecked if field in src['ignore_fields'] else Qt.CheckState.Checked

    def restore_defaults(self):
        self.beginResetModel()
        self.overrides = {f: self.state(f, True) for f in self.fields}
        self.endResetModel()

    def commit(self):
        ignored_fields = {x for x in self.prefs['ignore_fields'] if x not in self.overrides}
        changed = {k for k, v in self.overrides.items() if v == Qt.CheckState.Unchecked}
        self.prefs['ignore_fields'] = list(ignored_fields.union(changed))


# }}}


class FieldsList(QListView):
    def sizeHint(self):
        return self.minimumSizeHint()


class ConfigWidget(QWidget):
    def __init__(self, plugin):
        QWidget.__init__(self)
        self.plugin = plugin

        self.overl = l = QVBoxLayout(self)
        self.gb = QGroupBox(_('Metadata fields to download'), self)
        if plugin.config_help_message:
            self.pchm = QLabel(plugin.config_help_message)
            self.pchm.setWordWrap(True)
            self.pchm.setOpenExternalLinks(True)
            l.addWidget(self.pchm, 10)
        l.addWidget(self.gb)
        g = QVBoxLayout(self.gb)
        g.setContentsMargins(0, 0, 0, 0)
        self.fields_view = v = FieldsList(self)
        g.addWidget(v)
        v.setFlow(QListView.Flow.LeftToRight)
        v.setWrapping(True)
        v.setResizeMode(QListView.ResizeMode.Adjust)
        self.fields_model = FieldsModel(self.plugin)
        self.fields_model.initialize()
        v.setModel(self.fields_model)
        self.memory = []
        self.widgets = []
        self.l = QGridLayout()
        self.l.setContentsMargins(0, 0, 0, 0)
        l.addLayout(self.l, 100)
        for opt in plugin.options:
            self.create_widgets(opt)

    def create_widgets(self, opt):
        val = self.plugin.prefs[opt.name]
        if opt.type == 'number':
            c = QSpinBox if isinstance(opt.default, numbers.Integral) else QDoubleSpinBox
            widget = c(self)
            widget.setRange(int(min(widget.minimum(), 20 * val)), int(max(widget.maximum(), 20 * val)))
            widget.setValue(val)
        elif opt.type == 'string':
            widget = QLineEdit(self)
            widget.setText(val or '')
        elif opt.type == 'bool':
            widget = QCheckBox(opt.label, self)
            widget.setChecked(bool(val))
        elif opt.type == 'choices':
            widget = QComboBox(self)
            items = list(opt.choices.items())
            items.sort(key=lambda k_v: sort_key(k_v[1]))
            for key, label in items:
                widget.addItem(label, (key))
            idx = widget.findData(val)
            widget.setCurrentIndex(idx)
        widget.setToolTip(textwrap.fill(opt.desc))
        setattr(widget, 'opt', opt)
        self.widgets.append(widget)
        r = self.l.rowCount()
        if opt.type == 'bool':
            self.l.addWidget(widget, r, 0, 1, self.l.columnCount())
        else:
            l = QLabel(opt.label)
            l.setToolTip(widget.toolTip())
            self.memory.append(l)
            l.setBuddy(widget)
            self.l.addWidget(l, r, 0, 1, 1)
            self.l.addWidget(widget, r, 1, 1, 1)

    def commit(self):
        self.fields_model.commit()
        for w in self.widgets:
            if isinstance(w, (QSpinBox, QDoubleSpinBox)):
                val = w.value()
            elif isinstance(w, QLineEdit):
                val = str(w.text())
            elif isinstance(w, QCheckBox):
                val = w.isChecked()
            elif isinstance(w, QComboBox):
                idx = w.currentIndex()
                val = str(w.itemData(idx) or '')
            self.plugin.prefs[w.opt.name] = val



#!/usr/bin/env python
# License: GPLv3 Copyright: 2012, Kovid Goyal <kovid at kovidgoyal.net>

import builtins
import datetime
import json
import os
import posixpath
import sys
import traceback
from collections.abc import Sequence
from io import BytesIO
from typing import TYPE_CHECKING, NamedTuple

from calibre import prints
from calibre.constants import iswindows, numeric_version
from calibre.devices.errors import PathError
from calibre.devices.mtp.base import debug
from calibre.devices.mtp.defaults import DeviceDefaults
from calibre.devices.mtp.filesystem_cache import FileOrFolder, convert_timestamp
from calibre.ptempfile import PersistentTemporaryDirectory, SpooledTemporaryFile
from calibre.utils.filenames import shorten_components_to
from calibre.utils.icu import lower as icu_lower
from calibre.utils.localization import _
from polyglot.builtins import as_bytes

if TYPE_CHECKING:
    from calibre.devices.mtp.windows.driver import MTP_DEVICE as BASE
else:  # noqa
    if iswindows:
        from calibre.devices.mtp.windows.driver import MTP_DEVICE as BASE
    else:
        from calibre.devices.mtp.unix.driver import MTP_DEVICE as BASE
DEFAULT_THUMBNAIL_HEIGHT = 320


class MTPInvalidSendPathError(PathError):
    def __init__(self, folder):
        PathError.__init__(self, f'Trying to send to ignored folder: {folder}')
        self.folder = folder


class ListEntry(NamedTuple):
    name: str
    is_folder: bool
    size: int
    mtime: datetime.datetime


class MTP_DEVICE(BASE):
    METADATA_CACHE = 'metadata.calibre'
    DRIVEINFO = 'driveinfo.calibre'
    NEWS_IN_FOLDER = True
    MAX_PATH_LEN = 230
    THUMBNAIL_HEIGHT = DEFAULT_THUMBNAIL_HEIGHT
    CAN_SET_METADATA = []
    BACKLOADING_ERROR_MESSAGE = None
    MANAGES_DEVICE_PRESENCE = True
    FORMATS = ['epub', 'azw3', 'mobi', 'pdf']
    DEVICE_PLUGBOARD_NAME = 'MTP_DEVICE'
    SLOW_DRIVEINFO = True
    ASK_TO_ALLOW_CONNECT = True

    def __init__(self, *args, **kwargs):
        BASE.__init__(self, *args, **kwargs)
        self.plugboards = self.plugboard_func = None
        self._prefs = None
        self.device_defaults = DeviceDefaults()
        self.current_device_defaults = {}
        self.current_vid = self.current_pid = -1
        self.calibre_file_paths = {'metadata': self.METADATA_CACHE, 'driveinfo': self.DRIVEINFO}
        self.highlight_ignored_folders = False

    @property
    def prefs(self):
        from calibre.utils.config import JSONConfig

        if self._prefs is None:
            self._prefs = p = JSONConfig('mtp_devices')
            p.defaults['format_map'] = self.FORMATS
            p.defaults['send_to'] = [
                'Calibre_Companion',
                'Books',
                'eBooks/import',
                'eBooks',
                'wordplayer/calibretransfer',
                'sdcard/ebooks',
                'Android/data/com.amazon.kindle/files',
                'kindle',
                'NOOK',
                'Documents',
            ]
            p.defaults['send_template'] = '{title} - {authors}'
            p.defaults['blacklist'] = []
            p.defaults['history'] = {}
            p.defaults['rules'] = []
            p.defaults['ignored_folders'] = {}
        return self._prefs

    @property
    def is_kindle(self) -> bool:
        return self.current_vid == 0x1949

    def is_folder_ignored(self, storage_or_storage_id, path, ignored_folders=None):
        storage_id = str(getattr(storage_or_storage_id, 'object_id', storage_or_storage_id))
        lpath = tuple(icu_lower(name) for name in path)
        if self.is_kindle and lpath and lpath[-1].endswith('.sdr'):
            return True
        if ignored_folders is None:
            ignored_folders = self.get_pref('ignored_folders')
        if storage_id in ignored_folders:
            # Use the users ignored folders settings
            return '/'.join(lpath) in {icu_lower(x) for x in ignored_folders[storage_id]}

        # Implement the default ignore policy

        # Top level ignores
        if lpath[0] in {
            'alarms',
            'dcim',
            'movies',
            'music',
            'notifications',
            'screenshots',
            'pictures',
            'ringtones',
            'samsung',
            'sony',
            'htc',
            'bluetooth',
            'fonts',
            'games',
            'lost.dir',
            'video',
            'whatsapp',
            'image',
            'com.zinio.mobile.android.reader',
        }:
            return True
        if lpath[0].startswith('.') and lpath[0] not in ('.tolino', '.notebooks'):
            # apparently the Tolino for some reason uses a hidden folder for its library, sigh.
            # Kindle Scribe stores user notebooks in subfolders of '.notebooks'
            return True
        if lpath[0] == 'system' and not self.is_kindle:
            # on Kindles we need the system/thumbnails folder for the amazon cover bug workaround
            return True

        if len(lpath) > 1:
            if lpath[0] == 'android':
                # Ignore everything in Android apart from a few select folders
                if lpath[1] != 'data':
                    return True
                if len(lpath) > 2 and lpath[2] != 'com.amazon.kindle':
                    return True
            elif lpath[0] == 'system':
                if lpath[1] != 'thumbnails':
                    return True
        return False

    def configure_for_kindle_app(self):
        proxy = self.prefs
        with proxy:
            proxy['format_map'] = ['azw3', 'mobi', 'azw', 'azw1', 'azw4', 'pdf']
            proxy['send_template'] = '{title} - {authors}'
            orig = list(proxy['send_to'])
            for folder in ('kindle', 'Android/data/com.amazon.kindle/files'):
                if folder in orig:
                    orig.remove(folder)
                orig.insert(0, folder)
            proxy['send_to'] = orig

    def configure_for_generic_epub_app(self):
        with self.prefs:
            for x in ('format_map', 'send_template', 'send_to'):
                del self.prefs[x]

    def open(self, connected_device, library_uuid):
        from calibre.utils.date import isoformat, utcnow

        device = connected_device
        self.current_library_uuid = library_uuid
        self.location_paths = None
        self.driveinfo = {}
        BASE.open(self, device, library_uuid)
        h = self.prefs['history']
        if self.current_serial_num:
            h[self.current_serial_num] = (self.current_friendly_name, isoformat(utcnow()))
            self.prefs['history'] = h

        self.current_device_defaults, self.current_vid, self.current_pid = self.device_defaults(device, self)
        self.calibre_file_paths = self.current_device_defaults.get('calibre_file_paths', {'metadata': self.METADATA_CACHE, 'driveinfo': self.DRIVEINFO})
        self.THUMBNAIL_HEIGHT = DEFAULT_THUMBNAIL_HEIGHT
        if self.is_kindle:
            self.THUMBNAIL_HEIGHT = 500  # see kindle/driver.py
            try:
                self.sync_kindle_thumbnails()
            except Exception:
                import traceback

                traceback.print_exc()

    def list(self, path, recurse=False):
        if path.startswith('/'):
            q = self._main_id
            path = path[1:]
        elif path.startswith('card:/'):
            q = self._carda_id
            path = path[6:]
        for storage in self.filesystem_cache.entries:
            if storage.storage_id == q:
                if path:
                    path = path.replace(os.sep, '/')
                    parts = path.split('/')
                    if parts:
                        storage = storage.find_path(parts)
                        if storage is None:
                            return []
                return list(storage.list(recurse))
        return []

    def get_device_uid(self):
        return self.current_serial_num

    def ignore_connected_device(self, uid):
        bl = self.prefs['blacklist']
        if uid not in bl:
            bl.append(uid)
            self.prefs['blacklist'] = bl
        if self.is_mtp_device_connected:
            self.eject()

    def find_calibre_file_path(self, storage, key) -> builtins.list[str]:
        paths = self.calibre_file_paths[key]
        if isinstance(paths, str):
            paths = (paths,)
        if len(paths) == 1:
            return paths[0].split('/')
        # Find the first path whose parent directory exists
        for candidate in paths:
            path = candidate.split('/')
            if len(path) == 1:
                return path
            if storage.find_path(path[:-1]) is not None:
                return path
        return paths[0].split('/')

    def put_calibre_file(self, storage, key, stream, size):
        path = self.find_calibre_file_path(storage, key)
        parent = self.ensure_parent(storage, path)
        self.put_file(parent, path[-1], stream, size)

    # Device information {{{
    def _update_drive_info(self, storage, location_code, name=None):
        import uuid

        from calibre.utils.config import from_json, to_json
        from calibre.utils.date import isoformat, now

        f = storage.find_path(self.find_calibre_file_path(storage, 'driveinfo'))
        dinfo = {}
        if f is not None:
            try:
                stream = self.get_mtp_file(f)
                dinfo = json.load(stream, object_hook=from_json)
            except Exception:
                prints('Failed to load existing driveinfo.calibre file, with error:')
                traceback.print_exc()
                dinfo = {}
        if dinfo.get('device_store_uuid', None) is None:
            dinfo['device_store_uuid'] = str(uuid.uuid4())
        if dinfo.get('device_name', None) is None:
            dinfo['device_name'] = self.current_friendly_name
        if name is not None:
            dinfo['device_name'] = name
        dinfo['location_code'] = location_code
        dinfo['last_library_uuid'] = getattr(self, 'current_library_uuid', None)
        dinfo['calibre_version'] = '.'.join(str(i) for i in numeric_version)
        dinfo['date_last_connected'] = isoformat(now())
        dinfo['mtp_prefix'] = storage.storage_prefix
        raw = as_bytes(json.dumps(dinfo, default=to_json))
        self.put_calibre_file(storage, 'driveinfo', BytesIO(raw), len(raw))
        self.driveinfo[location_code] = dinfo

    def get_driveinfo(self):
        if not self.driveinfo:
            self.driveinfo = {}
            for sid, location_code in ((self._main_id, 'main'), (self._carda_id, 'A'), (self._cardb_id, 'B')):
                if sid is None:
                    continue
                self._update_drive_info(self.filesystem_cache.storage(sid), location_code)
        return self.driveinfo

    def get_device_information(self, end_session=True):
        self.report_progress(1.0, _('Get device information...'))
        dinfo = self.get_basic_device_information()
        return tuple(list(dinfo) + [self.driveinfo])

    def card_prefix(self, end_session=True):
        return (self._carda_id, self._cardb_id)

    def set_driveinfo_name(self, location_code, name):
        sid = {'main': self._main_id, 'A': self._carda_id, 'B': self._cardb_id}.get(location_code, None)
        if sid is None:
            return
        self._update_drive_info(self.filesystem_cache.storage(sid), location_code, name=name)

    # }}}

    # Get list of books from device, with metadata {{{
    def filesystem_callback(self, msg):
        self.report_progress(0, msg)

    def books(self, oncard=None, end_session=True):
        from calibre.devices.mtp.books import Book, BookList, JSONCodec

        self.report_progress(0, _('Listing files, this can take a while'))
        self.get_driveinfo()  # Ensure driveinfo is loaded
        sid = {'carda': self._carda_id, 'cardb': self._cardb_id}.get(oncard, self._main_id)
        if sid is None:
            return BookList(None)

        bl = BookList(sid)
        # If True then there is a mismatch between the ebooks on the device and
        # the metadata cache
        need_sync = False
        all_books = list(self.filesystem_cache.iterebooks(sid))
        steps = len(all_books) + 2
        count = 0

        self.report_progress(0, _('Reading e-book metadata'))
        # Read the cache if it exists
        storage = self.filesystem_cache.storage(sid)
        cache = storage.find_path(self.find_calibre_file_path(storage, 'metadata'))
        if cache is not None:
            json_codec = JSONCodec()
            try:
                stream = self.get_mtp_file(cache)
                json_codec.decode_from_file(stream, bl, Book, sid)
            except Exception:
                need_sync = True

        relpath_cache = {b.mtp_relpath: i for i, b in enumerate(bl)}

        for mtp_file in all_books:
            count += 1
            relpath = mtp_file.mtp_relpath
            idx = relpath_cache.get(relpath, None)
            if idx is not None:
                cached_metadata = bl[idx]
                del relpath_cache[relpath]
                if cached_metadata.size == mtp_file.size:
                    cached_metadata.datetime = mtp_file.last_modified.timetuple()
                    cached_metadata.path = mtp_file.mtp_id_path
                    debug('Using cached metadata for', '/'.join(mtp_file.full_path))
                    continue  # No need to update metadata
                book = cached_metadata
            else:
                book = Book(sid, '/'.join(relpath))
                bl.append(book)

            need_sync = True
            self.report_progress(count / steps, _('Reading metadata from %s') % ('/'.join(relpath)))
            try:
                book.smart_update(self.read_file_metadata(mtp_file))
                debug('Read metadata for', '/'.join(mtp_file.full_path))
            except Exception:
                prints('Failed to read metadata from', '/'.join(mtp_file.full_path))
                traceback.print_exc()
            book.size = mtp_file.size
            book.datetime = mtp_file.last_modified.timetuple()
            book.path = mtp_file.mtp_id_path

        # Remove books in the cache that no longer exist
        for idx in sorted(relpath_cache.values(), reverse=True):
            del bl[idx]
            need_sync = True

        if need_sync:
            self.report_progress(count / steps, _('Updating metadata cache on device'))
            self.write_metadata_cache(storage, bl)
        self.report_progress(1, _('Finished reading metadata from device'))
        return bl

    def read_file_metadata(self, mtp_file):
        from calibre.customize.ui import quick_metadata
        from calibre.ebooks.metadata.meta import get_metadata

        ext = mtp_file.name.rpartition('.')[-1].lower()
        if self.is_kindle and ext == 'kfx' and mtp_file.name != 'metadata.kfx':
            # locate the actual file containing KFX book metadata
            stream = BytesIO()
            try:
                self.get_file_by_name(stream, mtp_file.parent, *(mtp_file.name[:-4] + '.sdr', 'assets', 'metadata.kfx'))
            except Exception:
                pass  # expect failure for sideloaded books and personal documents
            else:
                stream.name = 'metadata.kfx'
                with quick_metadata:
                    metadata = get_metadata(stream, stream_type=ext, force_read_metadata=True, pattern=self.build_template_regexp())
                if metadata.title not in {'metadata', 'Unknown'}:
                    return metadata
        stream = self.get_mtp_file(mtp_file)
        with quick_metadata:
            return get_metadata(stream, stream_type=ext, force_read_metadata=True, pattern=self.build_template_regexp())

    def write_metadata_cache(self, storage, bl):
        from calibre.devices.mtp.books import JSONCodec

        if bl.storage_id != storage.storage_id:
            # Just a sanity check, should never happen
            return

        json_codec = JSONCodec()
        stream = SpooledTemporaryFile(10 * (1024**2))
        json_codec.encode_to_file(stream, bl)
        size = stream.tell()
        stream.seek(0)
        self.put_calibre_file(storage, 'metadata', stream, size)

    def sync_booklists(self, booklists, end_session=True):
        debug('sync_booklists() called')
        for bl in booklists:
            if getattr(bl, 'storage_id', None) is None:
                continue
            storage = self.filesystem_cache.storage(bl.storage_id)
            if storage is None:
                continue
            self.write_metadata_cache(storage, bl)
        debug('sync_booklists() ended')

    # }}}

    # Get files from the device {{{
    def get_file(self, path, outfile, end_session=True):
        f = self.filesystem_cache.resolve_mtp_id_path(path)
        self.get_mtp_file(f, outfile)

    def get_file_by_name(self, outfile, parent: FileOrFolder, *names: str) -> None:
        'Get the file parent/ + "/".join(names) and put it into outfile. Works with files not cached in FilesystemCache.'
        self.get_mtp_file_by_name(parent, *names, stream=outfile)

    def list_folder_by_name(self, parent: FileOrFolder, *names: str) -> tuple[ListEntry, ...]:
        'List the contents of the folder parent/ + "/".join(names). Works with folders not cached in FilesystemCache.'
        return tuple(ListEntry(x['name'], x['is_folder'], x['size'], convert_timestamp(x['modified'])) for x in self.list_mtp_folder_by_name(parent, *names))

    def prepare_addable_books(self, paths):
        tdir = PersistentTemporaryDirectory('_prepare_mtp')
        ans = []
        for path in paths:
            try:
                f = self.filesystem_cache.resolve_mtp_id_path(path)
            except Exception as e:
                ans.append((path, e, traceback.format_exc()))
                continue
            base = os.path.join(tdir, f'{f.object_id}')
            os.mkdir(base)
            name = f.name
            if iswindows:
                plen = len(base)
                max_len = 225 if self.is_kindle and path.endswith('.kfx') else 245  # allow for len of additional files
                name = ''.join(shorten_components_to(max_len - plen, [name]))
            with open(os.path.join(base, name), 'wb') as out:
                try:
                    self.get_mtp_file(f, out)
                except Exception as e:
                    ans.append((path, e, traceback.format_exc()))
                else:
                    ans.append(out.name)
                    if self.is_kindle and path.endswith('.kfx'):
                        # copy additional KFX files from the associated .sdr folder
                        try:
                            out_folder, out_file = os.path.split(out.name)
                            new_sdr_path = os.path.join(out_folder, out_file[:-4] + '.sdr')
                            os.mkdir(new_sdr_path)
                            self.scan_sdr_for_kfx_files(f.parent, (f.name[:-4] + '.sdr', 'assets'), new_sdr_path)
                        except Exception:
                            traceback.print_exc()
        return ans

    def scan_sdr_for_kfx_files(self, parent, folders, new_sdr_path):
        for ff in self.list_folder_by_name(parent, *folders):
            if ff.is_folder:
                self.scan_sdr_for_kfx_files(parent, folders + (ff.name,), new_sdr_path)
            elif ff.name.endswith('.kfx') or ff.name == 'voucher':
                name = ''.join(shorten_components_to(245 - len(new_sdr_path), [ff.name])) if iswindows else ff.name
                with open(os.path.join(new_sdr_path, name), 'wb') as out:
                    self.get_file_by_name(out, parent, *(folders + (ff.name,)))

    # }}}

    # Sending files to the device {{{

    def set_plugboards(self, plugboards, pb_func):
        self.plugboards = plugboards
        self.plugboard_func = pb_func

    def create_upload_path(self, path, mdata, fname, routing):
        from calibre.devices.utils import create_upload_path
        from calibre.utils.filenames import ascii_filename as sanitize

        ext = fname.rpartition('.')[-1].lower()
        path = routing.get(ext, path)

        filepath = create_upload_path(
            mdata,
            fname,
            self.save_template,
            sanitize,
            prefix_path=path,
            path_type=posixpath,
            maxlen=self.MAX_PATH_LEN,
            use_subdirs='/' in self.save_template,
            news_in_folder=self.NEWS_IN_FOLDER,
        )
        return tuple(x for x in filepath.split('/'))

    def prefix_for_location(self, on_card):
        if self.location_paths is None:
            self.location_paths = {}
            for sid, loc in ((self._main_id, None), (self._carda_id, 'carda'), (self._cardb_id, 'cardb')):
                if sid is not None:
                    storage = self.filesystem_cache.storage(sid)
                    prefixes = self.get_pref('send_to')
                    p = None
                    for path in prefixes:
                        path = path.replace(os.sep, '/')
                        if storage.find_path(path.split('/')) is not None:
                            p = path
                            break
                    if p is None:
                        p = 'Books'
                    self.location_paths[loc] = p

        return self.location_paths[on_card]

    def ensure_parent(self, storage, path):
        parent = storage
        pos = list(path)[:-1]
        while pos:
            name = pos[0]
            pos = pos[1:]
            parent = self.create_folder(parent, name)
        return parent

    def upload_books(self, files, names, on_card=None, end_session=True, metadata=None):
        debug('upload_books() called')
        from calibre.devices.utils import sanity_check

        sanity_check(on_card, files, self.card_prefix(), self.free_space())
        prefix = self.prefix_for_location(on_card)
        sid = {'carda': self._carda_id, 'cardb': self._cardb_id}.get(on_card, self._main_id)
        bl_idx = {'carda': 1, 'cardb': 2}.get(on_card, 0)
        storage = self.filesystem_cache.storage(sid)

        ans = []
        self.report_progress(0, _('Transferring books to device...'))
        i, total = 0, len(files)

        routing = dict(self.get_pref('rules'))

        assert metadata is not None
        for infile, fname, mi in zip(files, names, metadata):
            path = self.create_upload_path(prefix, mi, fname, routing)
            if path and self.is_folder_ignored(storage, path):
                raise MTPInvalidSendPathError('/'.join(path))
            parent = self.ensure_parent(storage, path)
            if hasattr(infile, 'read'):
                pos = infile.tell()
                infile.seek(0, 2)
                sz = infile.tell()
                infile.seek(pos)
                stream = infile
                close = False
            else:
                sz = os.path.getsize(infile)
                stream = open(infile, 'rb')
                close = True
            relpath = parent.mtp_relpath + (path[-1].lower(),)
            try:
                mtp_file = self.put_file(parent, path[-1], stream, sz)
                try:
                    self.upload_cover(parent, relpath, storage, mi, stream)
                    # Upload the apnx file
                    if self.is_kindle:
                        from calibre.devices.kindle.driver import get_apnx_opts

                        apnx_opts = get_apnx_opts()
                        if apnx_opts.send_apnx:
                            self.upload_apnx(path, storage, mi, infile, apnx_opts)
                except Exception:
                    import traceback

                    traceback.print_exc()
            finally:
                if close:
                    stream.close()
            ans.append((mtp_file, bl_idx))
            i += 1
            self.report_progress(i / total, _('Transferred %s to device') % mi.title)

        self.report_progress(1, _('Transfer to device finished...'))
        debug('upload_books() ended')
        return ans

    def upload_cover(
        self,
        parent_folder: FileOrFolder,
        relpath_of_ebook_on_device: Sequence[str],
        storage: FileOrFolder,
        mi,
        ebook_file_as_stream,
    ):
        if self.is_kindle:
            self.upload_kindle_thumbnail(parent_folder, relpath_of_ebook_on_device, storage, mi, ebook_file_as_stream)

    # Kindle cover thumbnail handling {{{

    def upload_kindle_thumbnail(
        self,
        parent_folder: FileOrFolder,
        relpath_of_ebook_on_device: Sequence[str],
        storage: FileOrFolder,
        mi,
        ebook_file_as_stream,
    ):
        coverdata = getattr(mi, 'thumbnail', None)
        if not coverdata or not coverdata[2]:
            return
        from calibre.devices.kindle.driver import thumbnail_filename

        tfname = thumbnail_filename(ebook_file_as_stream)
        if not tfname:
            return
        thumbpath = 'system', 'thumbnails', tfname
        cover_stream = BytesIO(coverdata[2])
        sz = len(coverdata[2])
        try:
            parent = self.ensure_parent(storage, thumbpath)
        except Exception as err:
            print(f'Failed to upload cover thumbnail to system/thumbnails with error: {err}', file=sys.stderr)
            return
        self.put_file(parent, tfname, cover_stream, sz)
        cover_stream.seek(0)
        cache_path = 'amazon-cover-bug', tfname
        parent = self.ensure_parent(storage, cache_path)
        self.put_file(parent, tfname, cover_stream, sz)
        # mapping from ebook relpath to thumbnail filename
        from hashlib import sha1

        index_name = sha1('/'.join(relpath_of_ebook_on_device).encode()).hexdigest()
        data = tfname.encode()
        self.put_file(parent, index_name, BytesIO(data), len(data))

    def delete_kindle_cover_thumbnail_for(self, storage: FileOrFolder, mtp_relpath: Sequence[str]) -> None:
        from hashlib import sha1

        index_name = sha1('/'.join(mtp_relpath).encode()).hexdigest()
        index = storage.find_path(('amazon-cover-bug', index_name))
        if index is not None:
            data = BytesIO()
            self.get_mtp_file(index, data)
            tfname = data.getvalue().decode().strip()
            if tfname:
                thumbnail = storage.find_path(('system', 'thumbnails', tfname))
                if thumbnail is not None:
                    self.delete_file_or_folder(thumbnail)
                cache = storage.find_path(('amazon-cover-bug', tfname))
                if cache is not None:
                    self.delete_file_or_folder(cache)
                self.delete_file_or_folder(index)

    def sync_kindle_thumbnails(self):
        for storage in self.filesystem_cache.entries:
            self._sync_kindle_thumbnails(storage)

    def _sync_kindle_thumbnails(self, storage):
        system_thumbnails_dir = storage.find_path(('system', 'thumbnails'))
        amazon_cover_bug_cache_dir = storage.find_path(('amazon-cover-bug',))
        if system_thumbnails_dir is None or amazon_cover_bug_cache_dir is None:
            return
        debug('Syncing cover thumbnails to workaround amazon cover bug')
        system_thumbnails = {x.name: x for x in system_thumbnails_dir.files}
        count = 0
        for f in amazon_cover_bug_cache_dir.files:
            s = system_thumbnails.get(f.name)
            if s is not None and s.size != f.size:
                count += 1
                data = BytesIO()
                self.get_mtp_file(f, data)
                data.seek(0)
                sz = len(data.getvalue())
                self.put_file(system_thumbnails_dir, f.name, data, sz)
        debug(f'Restored {count} cover thumbnails that were destroyed by Amazon')

    # }}}

    def upload_apnx(self, path_of_book_on_device, storage, mi, filepath, apnx_opts):
        debug('upload_apnx() called')
        name = path_of_book_on_device[-1].rpartition('.')[0]
        debug('Uploading APNX file for', path_of_book_on_device)
        from calibre.devices.kindle.apnx import APNXBuilder
        from calibre.ptempfile import PersistentTemporaryFile

        apnx_local_file = PersistentTemporaryFile('.apnx')
        apnx_local_path = apnx_local_file.name
        apnx_local_file.close()

        try:
            custom_page_count = 0
            cust_col_name = apnx_opts.custom_col_name
            if cust_col_name:
                try:
                    custom_page_count = int(mi.get(cust_col_name, 0))
                except Exception:
                    pass

            method = apnx_opts.apnx_method

            cust_col_method = apnx_opts.method_col_name
            if cust_col_method:
                try:
                    method = str(mi.get(cust_col_method)).lower()
                    if method is not None:
                        method = method.lower()
                        if method not in ('fast', 'accurate', 'pagebreak'):
                            method = None
                except Exception:
                    prints(f'Invalid custom column method: {cust_col_method}, ignoring')

            apnx_builder = APNXBuilder()
            apnx_builder.write_apnx(filepath, apnx_local_path, method=method, page_count=custom_page_count)

            apnx_size = os.path.getsize(apnx_local_path)

            with open(apnx_local_path, 'rb') as apnx_stream:
                apnx_filename = f'{name}.apnx'
                apnx_path = tuple(path_of_book_on_device[:-1]) + (f'{name}.sdr', apnx_filename)
                sdr_parent = self.ensure_parent(storage, apnx_path)
                debug('Uploading APNX to:', apnx_path)
                self.put_file(sdr_parent, apnx_filename, apnx_stream, apnx_size)
        except Exception:
            print('Failed to generate APNX', file=sys.stderr)
            import traceback

            traceback.print_exc()
        finally:
            os.remove(apnx_local_path)
        debug('upload_apnx() ended')

    def add_books_to_metadata(self, locations, metadata, booklists):
        mtp_files = locations
        debug('add_books_to_metadata() called')
        from calibre.devices.mtp.books import Book

        i, total = 0, len(mtp_files)
        self.report_progress(0, _('Adding books to device metadata listing...'))
        for x, mi in zip(mtp_files, metadata):
            mtp_file, bl_idx = x
            bl = booklists[bl_idx]
            book = Book(mtp_file.storage_id, '/'.join(mtp_file.mtp_relpath), other=mi)
            book = bl.add_book(book, replace_metadata=True)
            if book is not None:
                book.size = mtp_file.size
                book.datetime = mtp_file.last_modified.timetuple()
                book.path = mtp_file.mtp_id_path
            i += 1
            self.report_progress(i / total, _('Added %s') % mi.title)

        self.report_progress(1, _('Adding complete'))
        debug('add_books_to_metadata() ended')

    # }}}

    # Removing books from the device {{{
    def recursive_delete(self, obj):
        parent = self.delete_file_or_folder(obj)
        if parent.empty and parent.can_delete and not parent.is_system:
            try:
                self.recursive_delete(parent)
            except Exception:
                prints('Failed to delete parent: {}, ignoring'.format('/'.join(parent.full_path)))

    def delete_books(self, paths, end_session=True):
        self.report_progress(0, _('Deleting books from device...'))

        for i, path in enumerate(paths):
            f = self.filesystem_cache.resolve_mtp_id_path(path)
            fpath = f.mtp_relpath
            storage = f.storage
            self.recursive_delete(f)
            if self.is_kindle:
                self.delete_kindle_cover_thumbnail_for(storage, fpath)
            self.report_progress((i + 1) / float(len(paths)), _('Deleted %s') % path)
        self.report_progress(1, _('All books deleted'))

    def remove_books_from_metadata(self, paths, booklists):
        self.report_progress(0, _('Removing books from metadata'))

        class NextPath(Exception):
            pass

        for i, path in enumerate(paths):
            try:
                for bl in booklists:
                    for book in bl:
                        if book.path == path:
                            bl.remove_book(book)
                            raise NextPath('')
            except NextPath:
                pass
            self.report_progress((i + 1) / len(paths), _('Removed %s') % path)

        self.report_progress(1, _('All books removed'))

    # }}}

    # Settings {{{

    def get_pref(self, key):
        """Get the setting named key. First looks for a device specific setting.
        If that is not found looks for a device default and if that is not
        found uses the global default."""
        dd = self.current_device_defaults if self.is_mtp_device_connected else {}
        dev_settings = self.prefs.get(f'device-{self.current_serial_num}', {})
        default_value = dd.get(key, self.prefs[key])
        return dev_settings.get(key, default_value)

    def config_widget(self):
        from calibre.gui2.device_drivers.mtp_config import MTPConfig

        return MTPConfig(self, highlight_ignored_folders=self.highlight_ignored_folders)

    def save_settings(self, config_widget):
        config_widget.commit()

    def settings(self):
        class Opts:
            def __init__(s):
                s.format_map = self.get_pref('format_map')

        return Opts()

    @property
    def save_template(self):
        return self.get_pref('send_template')

    def get_user_blacklisted_devices(self):
        bl = frozenset(self.prefs['blacklist'])
        ans = {}
        for dev, x in self.prefs['history'].items():
            name = x[0]
            if dev in bl:
                ans[dev] = name
        return ans

    def set_user_blacklisted_devices(self, devices):
        self.prefs['blacklist'] = list(devices)

    # }}}


def main():
    import io
    from pprint import pprint

    io
    dev = MTP_DEVICE(None)
    dev.startup()
    try:
        from calibre.devices.scanner import DeviceScanner

        scanner = DeviceScanner()
        scanner.scan()
        devs = scanner.devices
        cd = dev.detect_managed_devices(devs)
        if cd is None:
            raise ValueError('Failed to detect MTP device')
        dev.set_progress_reporter(prints)
        dev.open(cd, None)
        dev.filesystem_cache.dump()
        print(dev.device_debug_info(), flush=True)
        docs = dev.prefix_for_location(None)
        print('Prefix for main mem:', docs, flush=True)
        parent = dev.filesystem_cache.entries[0]
        entries = dev.list_folder_by_name(parent, docs)
        pprint(entries)
        pprint(dev.get_mtp_metadata_by_name(parent, docs, entries[0].name))
        files = [x for x in entries if not x.is_folder]
        f = io.BytesIO()
        dev.get_file_by_name(f, parent, docs, files[0].name)
        data = f.getvalue()
        print('Got', files[0].name, 'of size:', len(data))
        parent = parent.folder_named(docs)
        f.seek(0)
        destname = 'mtp-driver-test.' + files[0].name.rpartition('.')[2]
        m = dev.put_file(parent, destname, f, len(data))
        print('Put', destname, 'in the', docs, 'folder')
        pprint(m)
    except Exception:
        import traceback

        traceback.print_exc()
    finally:
        dev.shutdown()


if __name__ == '__main__':
    main()



# License: GPLv3 Copyright: 2009, John Schember <john@nachtimwald.com>

from calibre.ebooks.conversion.config import OPTIONS
from calibre.gui2.convert import QDoubleSpinBox, Widget
from calibre.gui2.convert.pdf_input_ui import Ui_Form
from calibre.utils.localization import _


class PluginWidget(Widget, Ui_Form):
    TITLE = _('PDF input')
    HELP = _('Options specific to') + ' PDF ' + _('input')
    COMMIT_NAME = 'pdf_input'
    ICON = 'mimetypes/pdf.png'

    def __init__(self, parent, get_option, get_help, db=None, book_id=None):
        Widget.__init__(self, parent, OPTIONS['input']['pdf'])
        self.db, self.book_id = db, book_id
        from calibre.ebooks.conversion.plugins.pdf_input import ENGINES

        self.opt_pdf_engine.addItems(ENGINES)
        self.initialize_options(get_option, get_help, db, book_id)
        self.opt_pdf_engine.currentIndexChanged.connect(self.update_engine_opts)
        self.update_engine_opts()

    def set_value_handler(self, g, val):
        if val is None and isinstance(g, QDoubleSpinBox):
            g.setValue(0.0)
            return True
        if g is self.opt_pdf_engine:
            idx = g.findText(val)
            if idx > -1:
                g.setCurrentIndex(idx)

    def update_engine_opts(self):
        enabled = self.opt_pdf_engine.currentText() == 'calibre'
        self.opt_pdf_footer_skip.setEnabled(enabled)
        self.opt_pdf_header_skip.setEnabled(enabled)
        self.opt_pdf_header_regex.setEnabled(enabled)
        self.opt_pdf_footer_regex.setEnabled(enabled)



#!/usr/bin/env python
# vim:fileencoding=utf-8
# License: GPLv3 Copyright: 2015, Kovid Goyal <kovid at kovidgoyal.net>
from __future__ import absolute_import, division, print_function, unicode_literals

store_version = 20  # Needed for dynamic plugin loading

from calibre.gui2.store import StorePlugin
from calibre.gui2.store.amazon_base import AmazonStore


class Base(AmazonStore):
    scraper_storage = []
    SEARCH_BASE_URL = 'https://www.amazon.com.au/s/'
    SEARCH_BASE_QUERY = {'url': 'search-alias=digital-text'}
    DETAILS_URL = 'https://amazon.com.au/dp/'
    STORE_LINK = 'https://www.amazon.com.au'


class AmazonKindleStore(Base, StorePlugin):
    pass


if __name__ == '__main__':
    Base().develop_plugin()
